import click
from oogway.fees import get_fees

@click.command()
@click.option('--timeframe', default='fastest', help="Timeframe for confirmation in blocks (fastest | 3 | 6)")
def fees(timeframe):
    fee = get_fees(timeframe=timeframe)
    fee = str(fee)
    click.echo('\n%s Satoshis/byte\n' % fee)



import click
from oogway.request import request_payment as reqp

@click.command()
@click.argument('address', nargs=1)
@click.argument('amount', nargs=1)
@click.option('--expire', default=0, help="payment request expiration in minutes")
@click.option('--message', default="", help="message")
def request_payment(address, amount, expire, message):
    request = reqp(address, amount, expire, message)
    click.echo('\n%s\n' % request)



from oogway.request import request_payment, parse_request
from oogway.validate import validate

REQ1_RES = "bitcoin:1FHXDkRLhoCziRjftaPB3fELUYrZomFanx?amount=0.00020000&time=1598319207&exp=3600&message=oogway%20requests"

REQ2_RES = "bitcoin:1FHXDkRLhoCziRjftaPB3fELUYrZomFanx?amount=0.00020000&time=1598319712"

def test_request_payment():
    req1 = request_payment("1FHXDkRLhoCziRjftaPB3fELUYrZomFanx", 20000, 60, "oogway requests")
    req2 = request_payment("1FHXDkRLhoCziRjftaPB3fELUYrZomFanx", 20000)

    assert req1[0:66] == REQ1_RES[0:66]
    assert req1[76:] == REQ1_RES[76:]
    assert req2[0:66] == REQ2_RES[0:66]

def test_parse_request():
    req = parse_request(REQ1_RES)
    address = req['address']
    amount = req['amount']
    created = req['created']
    status = req['status']

    assert validate.is_valid_address(address) == True
    assert type(amount) == str
    assert type(created) == str
    assert type(status) == str



from collections import deque
from hashlib import sha256

BASE58_ALPHABET = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
BASE58_ALPHABET_LIST = list(BASE58_ALPHABET)

def b58encode(bytestr):
    """encode to base58"""
    alphabet = BASE58_ALPHABET_LIST
    encoded = deque()
    append = encoded.appendleft
    _divmod = divmod

    num = int.from_bytes(bytestr, 'big')

    while num > 0:
        num, rem = _divmod(num, 58)
        append(alphabet[rem])
    
    encoded = ''.join(encoded)
    pad = 0
    for byte in bytestr:
        if byte == 0:
            pad += 1
        else:
            break
    
    return '1' * pad + encoded

def b58decode(bytestr, length):
    """decode from base58"""
    n = 0
    for char in bytestr:
        n = n * 58 + BASE58_ALPHABET.index(char)
    return n.to_bytes(length, 'big')

def b58check(bytestr):
    """check if base58 encoded address is valid"""
    try:
        bcbytes = b58decode(bytestr, 25)
        return bcbytes[-4:] == sha256(sha256(bcbytes[:-4]).digest()).digest()[:4]
    except Exception:
        return False



import decimal
from .convert import convert

class operation:

    @staticmethod
    def add_sats(*sats, unit="satoshi"):
        """addition operation on sats"""
        res = 0
        for sat in sats:
            if type(sat) != int:
                raise ValueError("Satoshi values must be integers")
            if sat <= 0:
                raise ValueError("Satoshi values must be positive")
            res += sat
        
        unit = unit.lower()
        if unit == 'satoshi':
            pass
        elif unit == 'btc':
            res = convert.to_btc(res, string=False)
        else:
            raise ValueError("Unit %s does not exist" % unit)

        return res

    @staticmethod
    def substract_sats(sat_a, sat_b, unit="satoshi"):
        """substraction operation on sats"""
        if type(sat_a) != int or type(sat_b) != int:
            raise ValueError("Satoshi values must be integers")
        
        if sat_a <= 0 or sat_b <= 0:
            raise ValueError("Satoshi values must be positive")
        
        if sat_b > sat_a:
            raise ValueError("Result cannot be negative")

        res = sat_a - sat_b

        unit = unit.lower()
        if unit == 'satoshi':
            pass
        elif unit == 'btc':
            res = convert.to_btc(res, string=False)
        else:
            raise ValueError("Unit %s does not exist" % unit)

        return res

    @staticmethod
    def add_btc(*btcs, unit="btc"):
        """addition operation on btc"""
        res = 0
        for btc in btcs:
            if type(btc) == int:
                pass
            elif type(btc) == float:
                pass
            elif type(btc) == decimal.Decimal:
                pass
            else:
                raise ValueError("BTC values must be numbers")
            if btc <= 0:
                raise ValueError("BTC values must be positive")
            res += btc
        
        unit = unit.lower()
        res = round(res, 8)
        res = decimal.Decimal(res)
        res = round(res, 8)
        if unit == 'btc':
            pass
        elif unit == 'satoshi':
            res = convert.to_satoshi(res, string=False)
        else:
            raise ValueError("Unit %s does not exist" % unit)

        return res

    @staticmethod
    def substract_btc(btc_a, btc_b, unit="btc"):
        """substraction operation on BTC"""
        if type(btc_a) == int:
            pass
        elif type(btc_a) == float:
            pass
        elif type(btc_a) == decimal.Decimal:
            pass
        else:
            raise ValueError("BTC values must be numbers")

        if type(btc_b) == int:
            pass
        elif type(btc_b) == float:
            pass
        elif type(btc_b) == decimal.Decimal:
            pass
        else:
            raise ValueError("BTC values must be numbers")
        
        if btc_a <= 0 or btc_b <= 0:
            raise ValueError("BTC values must be positive")
        
        if btc_b > btc_a:
            raise ValueError("Result cannot be negative")

        res = btc_a - btc_b

        unit = unit.lower()
        res = round(res, 8)
        res = decimal.Decimal(res)
        res = round(res, 8)
        if unit == 'btc':
            pass
        elif unit == 'satoshi':
            res = convert.to_satoshi(res, string=False)
        else:
            raise ValueError("Unit %s does not exist" % unit)

        return res



import pytest
from oogway.seed.gen import Mnemonic

def gen_phrases(strength=128, _range=5):
    """generates mnemonic phrases"""
    mn = Mnemonic()
    phrases = []
    for _ in range(_range):
        phrase = mn.generate(strength=strength)
        phrases.append(phrase)
    return phrases

def wordlist():
    mn = Mnemonic()
    return mn.wordlist

# 12 words mnemonic
phrases_128 = gen_phrases()
# 24 words mnemonic
phrases_256 = gen_phrases(strength=256)

@pytest.mark.parametrize("phrase", phrases_128)
def test_mnemonic_length_128(phrase):
    """validate mnemonic size of 12 words"""
    phrase_len = len(phrase.split())
    assert phrase_len == 12

@pytest.mark.parametrize("phrase", phrases_256)
def test_mnemonic_length_256(phrase):
    """validate mnemonic size of 24 words"""
    phrase_len = len(phrase.split())
    assert phrase_len == 24

@pytest.mark.parametrize("phrase", phrases_256)
def test_mnemonic_wordlist(phrase):
    """verify if mnemonic words are from wordlist"""
    _wordlist = wordlist()
    words = phrase.split()
    passes = []
    for word in words:
        if word in _wordlist:
            passes.append(True)
        else:
            passes.append(False)
    
    assert (False in passes) == False



import requests

fees_endpoint = "https://bitcoinfees.earn.com/api/v1/fees/recommended"
HARDCODED_FAST = 100
HARDCODED_THREE_BLOCKS = 95
HARDCODED_SIX_BLOCKS = 55

def get_fees(timeframe="fastest"):
    """get fees in satoshis per byte"""
    timeframe = timeframe.lower()
    if timeframe == "fastest":
        tf = "fastestFee"
        default_fee = HARDCODED_FAST
    elif timeframe == "3":
        tf = "halfHourFee"
        default_fee = HARDCODED_THREE_BLOCKS
    elif timeframe == "6":
        tf = "hourFee"
        default_fee = HARDCODED_SIX_BLOCKS
    else:
        raise ValueError("The specified timeframe %s does not exist" % timeframe)

    fee = requests.get(fees_endpoint)
    fee = fee.json()
    try:
        fee = fee[tf]
    except KeyError:
        fee = default_fee
    return fee



from oogway.convert import convert
import decimal

def test_to_satoshi():
    one_btc = convert.to_satoshi(amount=1, string=False)
    one_btc_str = convert.to_satoshi(amount=1, string=True)

    assert one_btc == int(1e8)
    assert type(one_btc) == int
    assert one_btc_str == str(int(1e8))
    assert type(one_btc_str) == str

def test_to_btc():
    one_btc_sats = convert.to_btc(amount=int(1e8), string=False)
    one_btc_sats_str = convert.to_btc(amount=int(1e8), string=True)

    assert one_btc_sats == decimal.Decimal('1.00000000')
    assert type(one_btc_sats) == decimal.Decimal
    assert one_btc_sats_str == '1.00000000'
    assert type(one_btc_sats_str) == str



import time
from urllib.parse import quote, urlparse, parse_qs
from .validate import validate
from .convert import convert

def request_payment(address, amount, exp=0, message=""):
    """form a payment request"""
    if validate.is_valid_address(address) == True:
        pass
    else:
        raise ValueError("Invalid Bitcoin address")
    
    btc = convert.to_btc(amount, string=True)

    now = time.time()
    now = int(now)
    now = str(now)
    
    if exp == 0:
        exp_str = ""
    else:
        exp = exp * 60
        exp = str(exp)
        exp_str = ("&exp=%s" % exp)

    if message == "":
        msg = ""
    else:
        if len(message) <= 120:
            message = quote(message)
            msg = ("&message=%s" % message)
        else:
            raise ValueError("Message must be smaller than 120 characters")

    req = ("bitcoin:%s?amount=%s&time=%s%s%s" % (address, btc, now, exp_str, msg))

    return req

def parse_request(uri):
    req = urlparse(uri)
    query = parse_qs(req.query)
    res = {}
    res['address'] = req.path
    res['amount'] = query['amount'][0]
    try:
        res['created'] = query['time'][0]
    except KeyError:
        res['created'] = 0

    res['status'] = "valid"

    try:
        res['exp'] = query['exp'][0]
        now = time.time()
        now = int(now)
        limit = int(res['created']) + int(res['exp'])
        if now > limit:
            res['status'] = "expired"
    except KeyError:
        pass

    try:
        res['message'] = query['message'][0]
    except KeyError:
        pass
    
    res['amount'] = convert.to_satoshi(res['amount'], string=True)
    
    if res['created'] != 0:
        created = int(res['created'])
        res['created'] = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(created))

    return res



from .utils.base58 import b58check
from .utils.bech32 import bech32_decode
import re

ADDRESS_REGEX = "^[13][a-km-zA-HJ-NP-Z1-9]{25,34}$"
P2PKH_REGEX = "^[1][a-km-zA-HJ-NP-Z1-9]{25,34}$"
P2SH_REGEX = "^[3][a-km-zA-HJ-NP-Z1-9]{25,34}$"
BECH32_REGEX = "^bc1[ac-hj-np-zAC-HJ-NP-Z02-9]{11,71}$"

def _bech32check(address):
    """decodes bech32 address to check validity"""
    a = bech32_decode(address)
    if a == (None, None):
        return False
    else:
        return True

def p2pkh_val(address):
    """check P2PKH address validity"""
    regex = re.match(P2PKH_REGEX, address)
    encoding = b58check(address)
    if regex and encoding:
        return True
    else:
        return False

def p2sh_val(address):
    """check P2SH address validity"""
    regex = re.match(P2SH_REGEX, address)
    encoding = b58check(address)
    if regex and encoding:
        return True
    else:
        return False

def bech32_val(address):
    """check Bech32 address validity"""
    regex = re.match(BECH32_REGEX, address)
    encoding = _bech32check(address)
    if regex and encoding:
        return True
    else:
        return False

class validate:
    """
    validate addresses by format
    or validate regardless of
    format (general)
    """
    @staticmethod
    def is_p2pkh(_address):
        """validate P2PKH addresses"""
        a = p2pkh_val(_address)
        return a
    
    @staticmethod
    def is_p2sh(_address):
        """validate P2SH addresses"""
        a = p2sh_val(_address)
        return a
    
    @staticmethod
    def is_bech32(_address):
        """validate Bech32 addresses"""
        a = bech32_val(_address)
        return a
    
    @staticmethod
    def is_valid_address(_address):
        """validate all addresses format"""
        p2pkh = p2pkh_val(_address)
        p2sh = p2sh_val(_address)
        bech32 = bech32_val(_address)

        if any([p2pkh, p2sh, bech32]):
            a = True
        else:
            a = False

        return a

    @staticmethod
    def address_format(_address):
        bech32 = bech32_val(_address)
        p2sh = p2sh_val(_address)
        p2pkh = p2pkh_val(_address)

        if bech32 == True:
            f = 'Bech32'
        elif p2sh == True:
            f = 'P2SH'
        elif p2pkh == True:
            f = 'P2PKH'
        else:
            raise ValueError("Invalid address")

        return f



from oogway.fees import get_fees

def test_get_fees():
    fastest = get_fees('fastest')
    three = get_fees('3')
    six = get_fees('6')

    assert type(fastest) == int
    assert type(three) == int
    assert type(six) == int



import click
from oogway.cli.cli_address import keypair, validate
from oogway.cli.cli_convert import convert_to_satoshi, convert_to_btc
from oogway.cli.cli_fees import fees
from oogway.cli.cli_request import request_payment

@click.group(invoke_without_command=True)
def cli():
    pass

cli.add_command(keypair)
cli.add_command(validate)
cli.add_command(convert_to_satoshi)
cli.add_command(convert_to_btc)
cli.add_command(fees)
cli.add_command(request_payment)



from .key import Key
from .validate import validate
from .convert import convert
from .operation import operation
from .fees import get_fees
from .request import request_payment, parse_request

name = "oogway"
__version__ = "0.3.2"
